import requests
from fuzzywuzzy import fuzz

from ..util.util import is_empty
from .util import MapKeys
from .util import MapUrls
from .util import error_baidu
from .util import fix_address
from .util import gcj2xx

KEY = MapKeys.baidu


def address_json(city: str, address: str, key=None):
  """百度地理编码接口"""
  if is_empty(address):
    return None
  if not key:
    key = KEY
  url = f'{MapUrls.baidu}?city={city}&address={address}&output=json&ak={key}&ret_coordtype=gcj02ll'
  js = requests.get(url).json()
  error_baidu(js)
  if js['status'] == 0:
    return js['result']
  return None


def place_json(city: str, keywords: str, key=None):
  """百度地点检索接口"""
  if is_empty(keywords):
    return None
  if not key:
    key = KEY
  url = f'{MapUrls.baidu_poi}?query={keywords}&region={city}&city_limit=true&output=json&ak={key}&ret_coordtype=gcj02ll'
  js = requests.get(url).json()
  error_baidu(js)
  if js['status'] == 0 and len(js['results']) >= 1:
    return js['results'][0]
  return None


def get_baidu(*,
              address,
              city,
              source,
              disable_cache=False,
              with_detail=True,
              key=None):
  """脉策geocode服务"""
  if is_empty(address):
    return None
  url = f'{MapUrls.mdt}?address={address}&city={city}&disable_cache={disable_cache}&with_detail={with_detail}&source={source}'
  req = requests.get(url)
  if req.status_code == 200:
    return req.json()['result'][0]['extra']
  elif req.status_code in (400, 403):
    if source == 'baidu':
      return address_json(city=city, address=address, key=key)
    elif source == 'baidu_poi':
      return place_json(city=city, keywords=address, key=key)
    else:
      raise ValueError('source参数错误')
  else:
    return None


def get_address_baidu(city: str, address: str, srs='wgs84', key=None):
  if is_empty(address):
    return {
      'rv': None,
      'score': 0,
      'lng': None,
      'lat': None,
    }
  city = city.rstrip('市')
  address = fix_address(address)
  res = get_baidu(city=city, address=address, source='baidu', key=key)
  if res:
    lnglat = [res['location']['lng'], res['location']['lat']]
    if res['precise'] == 1:
      score = 100
    else:
      score = res.get('comprehension') if res.get('comprehension') else res.get(
          'confidence')
    latlng = gcj2xx(lnglat, srs=srs)
    lng = latlng[1]
    lat = latlng[0]
  else:
    score, lng, lat = 0, None, None
  return {
    'rv': None,
    'score': score,
    'lng': lng,
    'lat': lat,
  }


def get_place_baidu(city: str, keywords: str, srs='wgs84', key=None):
  if is_empty(keywords):
    return {
      'rv': None,
      'score': 0,
      'lng': None,
      'lat': None,
    }
  city = city.rstrip('市')
  keywords = fix_address(keywords)
  res = get_baidu(city=city, address=keywords, source='baidu_poi', key=key)
  if res:
    rv = res['name']
    lnglat = [res['location']['lng'], res['location']['lat']]
    latlng = gcj2xx(lnglat, srs=srs)
    lng = latlng[1]
    lat = latlng[0]
    score = max(fuzz.ratio(rv, keywords.lstrip(city)),
                fuzz.partial_ratio(rv, keywords.lstrip(city)))
  else:
    rv, lng, lat, score = None, None, None, 0
  return {
    'rv': rv,
    'score': score,
    'lng': lng,
    'lat': lat,
  }



import csv

import pandas as pd


def to_csv_by_line(filename: str, data: list):
  """
  逐行写入csv文件
  :param filename: 文件名
  :param data: 要写入的数据列表
  :return:
  """
  with open(filename, 'a') as f:
    csv_write = csv.writer(f, dialect='excel')
    csv_write.writerow(data)


def to_sheets(df_dict: dict, filename: str):
  """
  将多个dataframe保存到不同的sheet中

  Args:
      df_dict: 要保存的数据集，格式为：{sheet_name: DataFrame}
      filename: 要保存的文件名
  """
  with pd.ExcelWriter(filename) as writer:
    for sheet_name, data in df_dict.items():
      data.to_excel(writer, sheet_name)



import re
from datetime import datetime

import pandas as pd


def auto2date(string, errors='raise'):
  if string in ['', None]:
    return None
  string = str(string)

  if '时' in string:
    _hour = re.findall('\d+时', string)[0]
    string = string.split(_hour)[0]

  if ('年' in string) & ('月' in string) & ('日' in string):
    return pd.to_datetime(string, format='%Y年%m月%d日')
  elif ('年' in string) & ('月' in string):
    return pd.to_datetime(string, format='%Y年%m月')
  elif '年' in string:
    return pd.to_datetime(string, format='%Y年')
  else:
    return pd.to_datetime(string, errors=errors)


def is_valid_date(string):
  """判断是否是一个有效的日期字符串"""
  try:
    auto2date(string)
    return True
  except:
    return False


def excel2date(dates, date_type='str'):
  """
  excel的数字样式时间格式转日期格式

  :param dates: 传入excel的日期格式，或ymd的日期格式
  :param date_type: 返回日期类型，str or date
  :return:
  """
  from xlrd import xldate_as_datetime

  if pd.isna(dates):
    return None

  dates = str(dates)
  if re.match('^\d{4,5}$', dates):
    _date = datetime.strftime(xldate_as_datetime(int(dates), 0), '%Y-%m-%d')
  elif is_valid_date(dates):
    _date = datetime.strftime(auto2date(dates), '%Y-%m-%d')
  else:
    _date = None

  if _date:
    if date_type in ('string', 'str', str):
      return _date
    elif date_type in ('date', 'datetime'):
      return datetime.strptime(_date, '%Y-%m-%d')
    else:
      raise ValueError('date_type参数错误，可选参数为str或date')
  else:
    return _date


class DT:
  # TODO(wangyukang): 补充日期方法
  @property
  def today(self):
    return

  @property
  def tomorrow(self):
    return

  @property
  def yesterday(self):
    return

  @property
  def the_day_after_tomorrow(self):
    return

  @property
  def the_day_before_yesterday(self):
    return

  @property
  def one_year_ago(self):
    return

  @property
  def d(self):
    return



from datetime import datetime

import pandas as pd

from ..util.geom import wkb_loads
from ..util.util import ensure_list
from ..util.util import fuzz_match
from ..util.util import list2dict
from ..util.util import to_float


def filter_valid_geom(df: pd.DataFrame,
                      c_geometry='geometry',
                      ignore_index=True):
  """仅保留有效的geometry，剔除空白和错误的geometry行"""
  if 'temp_xxx' in df.columns:
    raise KeyError('数据集中不能存在【temp_xxx】列')
  df['temp_xxx'] = df[c_geometry].apply(wkb_loads)
  df = df[df['temp_xxx'].notna()]
  if ignore_index:
    df = df.reset_index(drop=True)
  del df['temp_xxx']
  return df


def best_unique(df: pd.DataFrame,
                key_cols: (list, str),
                value_cols: (str, list) = None,
                filter=False,
                drop_if_null='all'):
  """
  优化的去重函数：
    为保证数据的完整性，去重时优先去除指定列中的空值

  :param df:
  :param key_cols: 按照哪些列去重
  :param value_cols: 优先去除那些列的空值，该列表是有顺序的
  :param filter:
  :param drop_if_null: 如何处理value_cols内值为空的列；'all'：都为空时删除该列，'any'：任意一列为空时就删除，None：保留空白
  :return:
  """
  key_cols = ensure_list(key_cols)
  if value_cols is None:
    value_cols = [i for i in df.columns if i not in key_cols]
  else:
    value_cols = ensure_list(value_cols)
  if drop_if_null is not None:
    df = df.dropna(subset=value_cols, how=drop_if_null).dropna(subset=key_cols,
                                                               how='all')
  df = df.sort_values(value_cols, na_position='first')
  df = df.drop_duplicates(key_cols, keep='last').reset_index(drop=True)
  if filter:
    df = df[key_cols + value_cols]
  return df


def table2dict(df: pd.DataFrame,
               key_col: str = None,
               value_col: (str, list) = None,
               orient: str = 'dict') -> dict:
  """
  DataFrame转字典

  :param df:
  :param key_col: 生成key的列
  :param value_col: 生成value的列
  :param orient: 生成dict的方式，默认'dict',还有 ‘list’, ‘series’, ‘split’, ‘records’, ‘index’
  :return:
  """
  if (key_col is None) or (value_col is None):
    cols = list(df.columns).copy()
    key_col = cols[0]
    value_col = cols[1]

  df = df[~df[key_col].isna()]
  df.set_index(key_col, inplace=True)

  if isinstance(value_col, list):
    df = df[value_col]
    return df.to_dict(orient=orient)
  else:
    df = df[[value_col]]
    return df.to_dict(orient=orient)[value_col]


def round_by_columns(df, col: list):
  """对整列进行四舍五入，默认绝对值大于1的数值保留两位小数，小于1 的保留4位"""

  def _round(x):
    if abs(x) >= 1:
      return round(x, 2)
    else:
      return round(x, 4)

  col = ensure_list(col)
  for i in col:
    df[i] = df[i].apply(lambda x: _round(x))
  return df


def standard(series: (pd.Series, list),
             q: float = 0.01,
             min_score: float = 0,
             minus: bool = False) -> (pd.Series, list):
  if minus:
    series = 1 / (series + 1)
  max_ = series.quantile(1 - q)
  min_ = series.quantile(q)
  series = series.apply(
      lambda x: (x - min_) / (max_ - min_) * (100 - min_score) + min_score)
  series[series >= 100] = 100
  series[series <= min_score] = min_score
  return series


def update_df(df: pd.DataFrame,
              new_df: pd.DataFrame,
              on: (str, list) = None,
              mode='update'):
  """
  根据某一列更新dataframe里的数据

  :param df: 待升级的
  :param new_df: 新表
  :param on: 根据哪一列升级,默认为None，使用index
  :param mode：处理方式，update：直接更新对应位置的数值，insert：只有对应位置为空时才更新
  :return:
  """
  v1 = len(df)
  if on is not None:
    on = ensure_list(on)
    new_df = new_df.drop_duplicates()
    if any(new_df[on].duplicated()):
      raise ValueError('new_df中有重复的索引列对应不同的值，请检查')
    new_df = df[on].drop_duplicates().merge(new_df, how='inner', on=on)
    df = df.set_index(on, drop=False)
    new_df = new_df.set_index(on, drop=False)
  if mode == 'update':
    df.update(new_df)
  elif mode == 'insert':
    df = df.combine_first(new_df)
  else:
    raise ValueError(f'参数{mode}错误,可选参数为 update or insert')
  df = df.reset_index(drop=True)
  if on is not None:
    if v1 != len(df):
      raise ValueError('update后Dataframe结构发生变化，请检查')
  return df


def date_to(series: pd.Series, mode: str = 'first') -> pd.Series:
  """
  将日期转为当月的第一天或最后一天

  :param series: pd.Serise
  :param mode: 'first' or 'last'
  :return:
  """
  from pandas.tseries.offsets import MonthEnd

  def trans(x):
    if x is not pd.NaT:
      y = int(x.year)
      m = int(x.month)
      d = int(x.day)
      return datetime(y, m, d)
    else:
      return None

  series = pd.to_datetime(series)
  if mode == 'first':
    series = series.apply(lambda x: x.replace(day=1))
  elif mode == 'last':
    series = pd.to_datetime(series, format="%Y%m") + MonthEnd(1)
  else:
    raise ValueError(f"{mode}不是正确的参数，请使用 'first' or 'last'")
  return series.apply(trans)


def fuzz_df(df: pd.DataFrame,
            col: str,
            target_serise: (list, pd.Series)) -> pd.DataFrame:
  """
  为DataFrame中的某一列，从某个集合中匹配相似度最高的元素

  :param df: 输入的dataframe
  :param col: 要匹配的列
  :param target_serise: 从何处匹配， list/pd.Serise
  :return:
  """
  df[[f'{col}_target',
      'normal_score',
      'partial_score']] = df.apply(lambda x: fuzz_match(x[col], target_serise),
                                   result_type='expand', axis=1)
  return df


def series_to_float(series: pd.Series, rex_method: str = 'mean') -> pd.Series:
  """
  pandas.Series: str --> float

  :param series: 要转换的pandas列
  :param rex_method: 计算mean,max,min， 默认为mean
  """
  return series.apply(lambda x: to_float(x, rex_method=rex_method))


def filter_by_df(df: pd.DataFrame, sizer: pd.DataFrame) -> pd.DataFrame:
  """根据一个dataframe筛选另一个dataframe"""
  sizer = sizer.drop_duplicates()
  cond = False
  for i in sizer.index:
    _cond = True
    for c in sizer:
      value = sizer[c][i]
      if pd.isna(value):
        _cond = _cond & df[c].isna()
      else:
        _cond = _cond & (df[c] == value)
    cond = cond | _cond
  return df[cond]


def expand_dict(df, c_src):
  """展开字典为多列"""
  return pd.concat(
      [
        df.drop(c_src, axis=1),
        df[c_src].apply(
            lambda x: pd.Series(x, dtype='object')
        ).set_index(df.index)
      ],
      axis=1
  )


def split_list_to_row(df, column):
  """将列表列中列表的元素拆成多行"""
  df[column] = df[column].apply(list2dict)
  return df.drop(columns=column).join(
      expand_dict(
          df[[column]], column
      ).stack(
      ).reset_index(
          level=1, drop=True
      ).rename(column)
  )


def dict2df(data: dict, c_key='key', c_value='value', as_index=False):
  """将字典转为dataframe"""
  df = pd.DataFrame()
  df[c_key] = data.keys()
  df[c_value] = data.values()
  if as_index:
    df.set_index(c_key, inplace=True)
  return df



import requests
from fuzzywuzzy import fuzz

from ..util.util import is_empty
from .util import MapKeys
from .util import MapUrls
from .util import error_amap
from .util import fix_address
from .util import gcj2xx

KEY = MapKeys.amap


def address_json(city: str, address: str, key=None):
  """高德地理编码接口"""
  if is_empty(address):
    return None
  if not key:
    key = KEY
  url = f'{MapUrls.amap}?address={address}&city={city}&key={key}'
  js = requests.get(url).json()
  error_amap(js)
  if js['status'] == '1' and int(js['count']) >= 1:
    return js['geocodes'][0]
  return None


def place_json(city: str, keywords: str, key=None):
  """高德地点检索接口"""
  if is_empty(keywords):
    return None
  if not key:
    key = KEY
  url = f'{MapUrls.amap_poi}?keywords={keywords}&city={city}&key={key}'
  js = requests.get(url).json()
  error_amap(js)
  if js['status'] == '1' and int(js['count']) >= 1:
    return js['pois'][0]
  return None


def get_amap(*,
             address,
             city,
             source,
             with_detail=True,
             disable_cache=False,
             key=None):
  """脉策geocode服务"""
  if is_empty(address):
    return None
  url = f'{MapUrls.mdt}?address={address}&city={city}&disable_cache={disable_cache}&with_detail={with_detail}&source={source}'
  req = requests.get(url)
  if req.status_code == 200:
    return req.json()['result'][0]['extra']
  elif req.status_code in (400, 403):
    if source == 'amap':
      return address_json(city=city, address=address, key=key)
    elif source == 'amap_poi':
      return place_json(city=city, keywords=address, key=key)
    else:
      raise ValueError('source参数错误')
  else:
    return None


def get_address_amap(city: str, address: str, srs: str = 'wgs84', key=None):
  if is_empty(address):
    return {
      'rv': None,
      'score': 0,
      'lng': None,
      'lat': None,
    }
  city = city.rstrip('市')
  address = fix_address(address).replace('|', '')
  address_dict = get_amap(city=city, address=address, source='amap', key=key)
  if address_dict:
    rv = address_dict['formatted_address']
    lnglat = address_dict['location'].split(',')
    latlng = gcj2xx(lnglat, srs=srs)
    lng = latlng[1]
    lat = latlng[0]
  else:
    rv, lng, lat = None, None, None
  return {
    'rv': rv,
    'score': 0,
    'lng': lng,
    'lat': lat,
  }


def get_place_amap(city: str, keywords: str, srs='wgs84', key=None):
  if is_empty(keywords):
    return {
      'rv': None,
      'score': 0,
      'lng': None,
      'lat': None,
    }
  city = city.rstrip('市')
  keywords = fix_address(keywords)
  res = get_amap(city=city, address=keywords, source='amap_poi', key=key)
  if res:
    rv = res['name']
    lnglat = res['location'].split(',')
    latlng = gcj2xx(lnglat, srs=srs)
    lng = latlng[1]
    lat = latlng[0]
    score = max(fuzz.ratio(rv, keywords.lstrip(city)),
                fuzz.partial_ratio(rv, keywords.lstrip(city)))
  else:
    rv, lng, lat, score = None, None, None, 0
  return {
    'rv': rv,
    'score': score,
    'lng': lng,
    'lat': lat,
  }



__version__ = '1.1.0'

from .etl import file
from .etl import load
from .etl import transformer
from .etl.extract import rdf
from .etl.transformer import expand_dict
from .geocode.geocode import geocode
from .geocode.geocode import geocode_df
from .util import coord_trans
from .util import dt
from .util import geom
from .util import id_number
from .util import os
from .util import phone_number
from .util import strings
from .util.geom import mark_tags_v2
from .util.geom import wkb_dumps
from .util.geom import wkb_loads
from .util.geom import wkt_dumps
from .util.geom import wkt_loads
from .util.os import ext
from .util.os import fn
from .util.os import mkdir_2
from .util.strings import drop_repeat_string
from .util.util import extract_num
from .util.util import is_empty
from .util.util import list2dict
from .util.util import not_empty
from .util.util import pinyin
from .util.util import segment



import os

import pandas as pd

from .etl.extract import rdf


class Dirs:
  p_neurond = '/Users/ricco/neurond'

  p_huamu_origin = '/Users/ricco/Project/collector_api_sdk/huamu'

  p_huamu_louyu = '/Users/ricco/Project/collector_api_sdk/huamu/花木街道-楼宇边界.csv'

  p_huamu_xiaoqu = '/Users/ricco/Project/collector_api_sdk/huamu/花木街道-小区边界.csv'

  p_huamu_shequ = '/Users/ricco/Project/collector_api_sdk/huamu/花木街道-社区边界.csv'

  p_huamu_juwei = '/Users/ricco/Project/collector_api_sdk/huamu/花木街道-居委边界.csv'

  p_huamu_jumin = '/Users/ricco/Project/collector_api_sdk/huamu/花木市民-居民基本信息表.csv'

  p_huamu_juzhu = '/Users/ricco/Project/collector_api_sdk/huamu/花木市民-居民居住信息.csv'

  p_huamu_fangwu = '/Users/ricco/Project/collector_api_sdk/huamu/花木市民-房屋基本信息表.csv'

  def p_street(self, city):
    return f'/Users/ricco/common_data/bd_region/{city}边界_街镇最新.csv'

  def p_region(self, city):
    return f'/Users/ricco/common_data/bd_region/{city}边界_街镇最新.csv'


class Datas(Dirs):
  def street(self, city='上海'):
    return rdf(self.p_street(city))

  def region(self, city='上海'):
    return rdf(self.p_region(city))

  def huamu_louyu(self):
    return rdf(self.p_huamu_louyu)

  def huamu_xiaoqu(self):
    return rdf(self.p_huamu_xiaoqu)

  def huamu_juwei(self):
    return rdf(self.p_huamu_juwei)

  def huamu_shequ(self):
    return rdf(self.p_huamu_shequ)

  def huamu_jumin(self):
    return rdf(self.p_huamu_jumin)

  def huamu_juzhu(self):
    return rdf(self.p_huamu_juzhu)

  def huamu_fangwu(self):
    return rdf(self.p_huamu_fangwu)


class AOI(Dirs):
  def building_geom(
      self,
      city: str,
      *,
      region: str = None,
      street: str = None,
      subfix: str = '百度',
      all_cols: bool = False,
  ):
    filename = f'{city}边界_建筑边界_{subfix}.parquet'
    fp = os.path.join('/Users/ricco/common_data/aoi/', filename)
    df = pd.read_parquet(fp)
    if not all_cols:
      df = df[['oid', '城市', '区县', '街镇', 'geometry']]
    if street:
      return df[df['街镇'] == street]
    if region:
      return df[df['区县'] == region]
    return df


class Ricco(Datas, AOI):
  """常用文件"""
  pass


Rc = Ricco()



import csv
import os
import warnings

import geopandas as gpd
import pandas as pd

from ..util.os import ext


def max_grid():
  """防止单个单元格文件过大而报错"""
  import sys
  maxint = sys.maxsize
  decrement = True
  while decrement:
    decrement = False
    try:
      csv.field_size_limit(maxint)
    except OverflowError:
      maxint = int(maxint / 10)
      decrement = True


def rdxls(
    filename,
    sheet_name=0,
    sheet_contains: str = None,
    errors='raise',
    dtype=None,
) -> pd.DataFrame:
  """
  读取excel文件

  :param filename: 文件名
  :param sheet_name: sheet表的名称
  :param sheet_contains: sheet表包含的字符串
  :param errors: 当没有对应sheet时，raise: 抛出错误, coerce: 返回空的dataframe
  :return:
  """
  if sheet_name == 0:
    if sheet_contains is not None:
      df = pd.read_excel(filename, sheet_name=None, dtype=dtype)
      sheet_list = [i for i in df.keys() if sheet_contains in i]
      if len(sheet_list) != 0:
        sheet_name = sheet_list[0]
        if len(sheet_list) == 1:
          print(f"sheet:  <'{sheet_name}'>")
        elif len(sheet_list) >= 2:
          warnings.warn(
              f"包含'{sheet_contains}'的sheet有{sheet_list}，所读取的sheet为:{sheet_name}")
        return df[sheet_name]
      else:
        if errors == 'coerce':
          warnings.warn(f'没有包含{sheet_contains}的sheet，请检查')
          return pd.DataFrame()
        elif errors == 'raise':
          raise ValueError(f'没有包含{sheet_contains}的sheet，请检查')
        else:
          raise KeyError("参数'error'错误, 可选参数为coerce和raise")
  else:
    print(f"sheet:  <'{sheet_name}'>")
  return pd.read_excel(filename, sheet_name=sheet_name, dtype=dtype)


def rdf(
    file_path: str,
    *,
    sheet_name=0,
    sheet_contains: str = None,
    encoding='utf-8-sig',
    info=False,
    dtype=None,
) -> pd.DataFrame:
  """
  常用文件读取函数，支持.csv/.xlsx/.shp/.parquet/.pickle/.feather'
  """
  max_grid()
  ex = ext(file_path)
  if ex == '.csv':
    try:
      df = pd.read_csv(file_path,
                       engine='python',
                       encoding=encoding,
                       dtype=dtype)
    except UnicodeDecodeError:
      df = pd.read_csv(file_path, engine='python', dtype=dtype)
  elif ex in ('.parquet', '.pickle', '.feather'):
    _t = ex.strip('.')
    df = getattr(pd, f'read_{_t}')(
        file_path
    )
  elif ex in ('.xls', '.xlsx'):
    df = rdxls(file_path,
               sheet_name=sheet_name,
               sheet_contains=sheet_contains,
               dtype=dtype)
  elif ex == '.shp':
    try:
      df = gpd.GeoDataFrame.from_file(file_path)
    except UnicodeEncodeError:
      df = gpd.GeoDataFrame.from_file(file_path, encoding='GBK')
  else:
    raise Exception('未知文件格式')
  if info:
    print(f'shape: {df.shape}')
    print(f'columns: {df.columns}')
  return df


def read_line_json(filename, encoding='utf-8') -> pd.DataFrame:
  """
  逐行读取json格式的文件

  :param filename: 文件名
  :param encoding: 文件编码，默认为utf-8
  :return:
  """
  import json
  records = []
  with open(filename, 'r', encoding=encoding) as file:
    for line in file:
      row = json.loads(line)
      records.append(row)
  return pd.DataFrame(records)


def bigdata2df(filename: str,
               chunksize: int = 10000,
               code: str = 'utf-8') -> pd.DataFrame:
  reader = pd.read_table(filename,
                         encoding=code,
                         sep=',',
                         skip_blank_lines=True,
                         iterator=True)
  loop = True
  chunks = []
  while loop:
    try:
      chunk = reader.get_chunk(chunksize)
      chunk.dropna(axis=0, inplace=True)
      chunks.append(chunk)
    except StopIteration:
      loop = False
      print('Iteration is stopped.')
  df = pd.concat(chunks, ignore_index=True, axis=1)
  return df


def read_parquet_by_dir(dir_path):
  df = pd.DataFrame()
  for filename in os.listdir(dir_path):
    if ext(filename) == '.parquet':
      df = pd.concat([df, rdf(os.path.join(dir_path, filename))])
  return df



import json
import logging
import warnings

import geojson
import geopandas as gpd
import numpy as np
import pandas as pd
from shapely import wkb
from shapely import wkt
from shapely.errors import GeometryTypeError
from shapely.errors import ShapelyDeprecationWarning
from shapely.errors import WKBReadingError
from shapely.geometry import LinearRing
from shapely.geometry import LineString
from shapely.geometry import MultiLineString
from shapely.geometry import MultiPoint
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon
from shapely.geometry import shape
from shapely.geos import WKTReadingError
from simplejson.errors import JSONDecodeError

from ..util.util import ensure_list
from ..util.util import first_notnull_value

warnings.filterwarnings('ignore', category=ShapelyDeprecationWarning)

GeomTypeSet = (
  Point, MultiPoint,
  Polygon, MultiPolygon,
  LineString, MultiLineString,
  LinearRing
)


class GeomFormat:
  wkb = 'wkb'
  wkt = 'wkt'
  shapely = 'shapely'
  geojson = 'geojson'
  unknown = 'unknown'


def crs_sh2000():
  """测绘院（上海2000）crs信息"""
  from ..resource.crs import CRS_SH2000
  return CRS_SH2000


def get_epsg_by_lng(lng):
  """通过经度获取epsg代码"""
  lng_epsg_mapping = {
    # 中央经线和epsg代码的对应关系
    75: 4534, 78: 4535, 81: 4536, 84: 4537, 87: 4538,
    90: 4539, 93: 4540, 96: 4541, 99: 4542,
    102: 4543, 105: 4544, 108: 4545, 111: 4546,
    114: 4547, 117: 4548, 120: 4549, 123: 4550,
    126: 4551, 129: 4552, 132: 4553, 135: 4554,
  }
  lng = np.median(lng)
  key = min(lng_epsg_mapping.keys(), key=lambda x: abs(x - lng))
  return lng_epsg_mapping.get(key)


def get_lng_by_city(city: str):
  from ..resource.epsg_code import CITY_POINT
  if city in CITY_POINT.keys():
    return CITY_POINT[city]['lng']
  else:
    city += '市'
    if city in CITY_POINT.keys():
      return CITY_POINT[city]['lng']
    else:
      warnings.warn(f'请补充{city}的epsg信息，默认返回经度120')
      return 120


def get_epsg(city: str):
  """根据城市查询epsg代码，用于投影"""
  return get_epsg_by_lng(get_lng_by_city(city))


def projection(
    gdf: gpd.GeoDataFrame,
    epsg: int = None,
    city: str = None):
  """投影变换"""
  if not epsg:
    if city:
      epsg = get_epsg(city)
    else:
      lngs = gdf['geometry'].centroid.x.tolist()
      epsg = get_epsg_by_lng(lngs)
  return gdf.to_crs(epsg=epsg)


def projection_lnglat(lnglat, crs_from, crs_to):
  from pyproj import Transformer
  transformer = Transformer.from_crs(crs_from, crs_to)
  return transformer.transform(xx=lnglat[1], yy=lnglat[0])


def wkb_loads(x, hex=True):
  warnings.filterwarnings('ignore',
                          'Geometry column does not contain geometry.',
                          UserWarning)

  if pd.isna(x):
    return None

  try:
    return wkb.loads(x, hex=hex)
  except (AttributeError, WKBReadingError):
    return None


def wkb_dumps(x, hex=True, srid=4326) -> (str, None):
  if pd.isna(x):
    return None

  try:
    return wkb.dumps(x, hex=hex, srid=srid)
  except AttributeError:
    return None


def wkt_loads(x):
  if pd.isna(x):
    return None
  try:
    return wkt.loads(x)
  except (AttributeError, WKTReadingError):
    return None


def wkt_dumps(x) -> (str, None):
  if pd.isna(x):
    return None
  try:
    return wkt.dumps(x)
  except AttributeError:
    return None


def geojson_loads(x):
  """geojson文本形式转为shapely格式"""
  if pd.isna(x):
    return None
  try:
    geom = shape(geojson.loads(x))
    if geom.is_empty:
      return None
    return geom
  except (JSONDecodeError, AttributeError, GeometryTypeError, TypeError):
    return None


def geojson_dumps(x) -> (str, None):
  """shapely转为geojson文本格式"""
  if pd.isna(x):
    return None
  try:
    geom = geojson.Feature(geometry=x)
    return json.dumps(geom.geometry)
  except TypeError:
    return None


def is_shapely(x, na=False) -> bool:
  """判断是否为shapely格式"""
  if pd.isna(x):
    return na
  if type(x) in GeomTypeSet:
    return True
  else:
    return False


def is_wkb(x, na=False) -> bool:
  """判断是否为wkb格式"""
  if pd.isna(x):
    return na
  try:
    wkb.loads(x, hex=True)
    return True
  except WKBReadingError:
    return False


def is_wkt(x, na=False) -> bool:
  """判断是否为wkt格式"""
  if pd.isna(x):
    return na
  try:
    wkt.loads(x)
    return True
  except WKTReadingError:
    return False


def is_geojson(x, na=False) -> bool:
  """判断是否为geojson格式"""
  if pd.isna(x):
    return na
  try:
    if shape(geojson.loads(x)).is_empty:
      return False
    return True
  except (JSONDecodeError, AttributeError, GeometryTypeError, TypeError):
    return False


def infer_geom_format(x):
  """推断geometry格式"""
  if is_shapely(x):
    return GeomFormat.shapely
  if is_wkb(x):
    return GeomFormat.wkb
  if is_wkt(x):
    return GeomFormat.wkt
  if is_geojson(x):
    return GeomFormat.geojson
  return GeomFormat.unknown


def ensure_multi_geom(geom):
  """将LineString和Polygon转为multi格式"""
  if geom.geom_type == 'LineString':
    return MultiLineString([geom])
  if geom.geom_type == 'Polygon':
    return MultiPolygon([geom])
  return geom


def multiline2multipolygon(multiline_shapely):
  """multiline转为multipolygon，直接首尾相连"""
  coords = []
  for line in multiline_shapely:
    lngs = line.xy[0]
    lats = line.xy[1]
    for i in range(len(lngs)):
      lng, lat = lngs[i], lats[i]
      point = Point((lng, lat))
      if len(coords) >= 1:
        if point != coords[-1]:
          coords.append(point)
      else:
        coords.append(point)
  return MultiPolygon([Polygon(coords)])


def geom_wkt2shapely(df, geometry='geometry',
                     epsg_code: int = 4326) -> gpd.GeoDataFrame:
  """wkt转gpd"""
  df[geometry] = df[geometry].apply(wkt_loads)
  return gpd.GeoDataFrame(df, geometry=geometry, crs=epsg_code)


def geom_wkb2lnglat(df, geometry='geometry', delete=False, within=False):
  """geometry转经纬度，求中心点经纬度"""
  df = geom_wkb2shapely(df, geometry=geometry)
  df = geom_shapely2lnglat(df, geometry=geometry, within=within, delete=delete)
  if not delete:
    df[geometry] = df[geometry].apply(wkb_dumps)
  return df


def get_inner_point(polygon: Polygon, within=True):
  """返回面内的一个点，默认返回中心点，当中心点不在面内则返回面内一个点"""
  if pd.isna(polygon):
    return None
  point = polygon.centroid
  if polygon.contains(point):
    return point
  else:
    if within:
      return polygon.representative_point()
    else:
      return point


def geom_wkt2wkb(df, geometry='geometry', epsg_code: int = 4326):
  """wkb转wkt"""
  df = geom_wkt2shapely(df, geometry=geometry, epsg_code=epsg_code)
  df[geometry] = df[geometry].apply(wkb_dumps)
  return df


def geom_wkb2shapely(df, geometry='geometry',
                     epsg_code: int = 4326) -> gpd.GeoDataFrame:
  """wkb直接转Dataframe"""
  df = df.copy()
  df[geometry] = df[geometry].apply(wkb_loads)
  return gpd.GeoDataFrame(df, geometry=geometry, crs=epsg_code)


def geom_lnglat2shapely(df, lng='lng', lat='lat', delete=True,
                        epsg_code: int = 4326) -> gpd.GeoDataFrame:
  """包含经纬度的DataFrame转GeoDataFrame"""
  from pandas.errors import SettingWithCopyWarning
  warnings.filterwarnings('ignore', category=SettingWithCopyWarning)
  df['geometry'] = df.apply(
      lambda d: Point((d[lng], d[lat])) if all([d[lng], d[lat]]) else None,
      axis=1
  )
  df = gpd.GeoDataFrame(df, crs=epsg_code)
  if delete:
    del df[lng], df[lat]
  return df


def geom_shapely2lnglat(df, geometry='geometry', within=False, delete=False):
  """
  shapely格式提取中心点转为经纬度。
  within: 范围的点是否再面内，默认False，直接返回中心点；
  当为True时，不在面内的中心点将用一个在面内的点代替
  """
  df['geometry_temp'] = df[geometry].apply(
      lambda x: get_inner_point(x, within=within) if x else None)
  df['lng'] = df['geometry_temp'].centroid.x
  df['lat'] = df['geometry_temp'].centroid.y
  del df['geometry_temp']
  if delete:
    del df[geometry]
  return df


def geom_lnglat2wkb(df, lng='lng', lat='lat', delete=False, code=4326):
  """经纬度转wkb格式的geometry"""
  df = geom_lnglat2shapely(df, 'lng', 'lat', delete=delete, epsg_code=code)
  df['geometry'] = df['geometry'].apply(wkb_dumps)
  if not delete:
    df = df.rename(columns={'lng': lng, 'lat': lat})
  return df


def geom_split_grids(df: gpd.GeoDataFrame, step: int, city: str = None):
  """
  根据所给边界划分固定边长的栅格

  Args:
      df: 边界文件，GeoDataFrame格式
      step: 栅格边长，单位：米
      city: 所属城市，用于投影

  Returns:

  """

  def get_xxyy(df):
    bounds_dict = df.bounds.T.to_dict()[0]
    minx = bounds_dict['minx']
    miny = bounds_dict['miny']
    maxx = bounds_dict['maxx']
    maxy = bounds_dict['maxy']
    xx = maxx - minx
    yy = maxy - miny
    return [xx, yy, minx, miny, maxx, maxy]

  def get_lnglat(df, i, j):
    _lng = df.loc[(df['i'] == i) &
                  (df['j'] == j), 'lng'].reset_index(drop=True)[0]
    _lat = df.loc[(df['i'] == i) &
                  (df['j'] == j), 'lat'].reset_index(drop=True)[0]
    return (_lng, _lat)

  def get_lnglat_sets(df, i, j):
    A = get_lnglat(df, i, j)
    B = get_lnglat(df, i, j + 1)
    C = get_lnglat(df, i + 1, j + 1)
    D = get_lnglat(df, i + 1, j)
    return Polygon((A, B, C, D, A))

  # 融合
  if not isinstance(df, gpd.GeoDataFrame):
    df = geom_wkb2shapely(df)
  df = df[['geometry']].dissolve()

  # 投影，获取实际的距离
  df_p = projection(df, city=city)

  # 分别获取投影前和投影后的点位及边长信息
  xy_o, xy_p = get_xxyy(df), get_xxyy(df_p)

  # 获取栅格的起止经纬度，步长等信息
  xr, yr = xy_o[0] / xy_p[0], xy_o[1] / xy_p[1]
  x_len, y_len = xr * step, yr * step
  x_num, y_num = int(xy_o[0] // x_len + 1), int(xy_o[1] // y_len + 1)
  logging.info(f'lng分段数：{x_num}，lat分段数：{y_num}')
  x_start, y_start = xy_o[2], xy_o[3]
  x_end, y_end = x_start + x_num * x_len, y_start + y_num * y_len

  # 构建经纬度列表
  lng_array, lat_array = np.linspace(x_start, x_end, num=x_num), np.linspace(
      y_start, y_end, num=y_num)
  lng_list, lat_list = list(lng_array), list(lat_array)

  # 构建栅格索引
  i_l, j_l, lng_l, lat_l = [], [], [], []
  for i, lng in enumerate(lng_list):
    for j, lat in enumerate(lat_list):
      i_l.append(i)
      j_l.append(j)
      lng_l.append(lng)
      lat_l.append(lat)
  df_temp = pd.DataFrame({'i': i_l, 'j': j_l, 'lng': lng_l, 'lat': lat_l})
  # 删除末尾的格子，避免后面出现out of range
  df_res = df_temp.loc[
    (df_temp['i'] != x_num - 1) & (df_temp['j'] != y_num - 1),
    ['i', 'j']].reset_index(drop=True)

  # 组合坐标集并转为Polygon
  df_res['geometry'] = df_res.apply(
      lambda df: get_lnglat_sets(df_temp, df['i'], df['j']), axis=1)

  # 生成grid_id列
  df_res['grid_id'] = df_res['i'].astype(str) + '-' + df_res['j'].astype(str)
  del df_res['i'], df_res['j']

  # 通过边界裁剪栅格
  df_res = gpd.GeoDataFrame(df_res, crs="epsg:4326")
  df.crs = 'epsg:4326'
  df_res = gpd.sjoin(df_res,
                     df,
                     how='inner',
                     predicate='intersects').drop('index_right', axis=1)

  # 将geometry转为wkb格式
  df_res['geometry'] = df_res['geometry'].apply(wkb_dumps)
  return df_res


def ensure_gdf(df, geometry='geometry'):
  geom = first_notnull_value(df[geometry])
  geom_format = infer_geom_format(geom)
  if geom_format == 'wkb':
    return geom_wkb2shapely(df, geometry=geometry)
  elif geom_format == 'wkt':
    return geom_wkt2shapely(df, geometry=geometry)
  elif geom_format == 'shapely':
    return gpd.GeoDataFrame(df, geometry=geometry)
  else:
    raise TypeError('未知的地理格式，支持wkb,wkt,shapely三种格式')


def mark_tags_v2(
    point_df: pd.DataFrame,
    polygon_df: pd.DataFrame,
    col_list: list = None,
    predicate='intersects',
    drop_geometry=False,
    geometry_format='wkb'):
  """
  使用面数据通过空间关联（sjoin）给点数据打标签

  Args:
      point_df: 点数据
      polygon_df: 面数据
      col_list: 面数据中要关联到结果中的列，若为空则全部关联
      predicate: 关联方法，默认'intersects'
      drop_geometry: 结果是否删除geometry，默认删除
      geometry_format: 输出的geometry格式，支持wkb,、wkt、shapely，默认wkb
  """
  if not col_list:
    col_list = polygon_df.columns.to_list()
  else:
    polygon_df = polygon_df[col_list + ['geometry']]

  col_list = ensure_list(col_list)
  for c in col_list:
    if c in point_df and c not in ['lng', 'lat', 'geometry']:
      c_n = f'{c}_origin'
      warnings.warn(f'点数据中存在面文件中待关联的列，已重命名：{c} --> {c_n}')
      point_df.rename(columns={c: c_n}, inplace=True)

  if 'geometry' in point_df:
    point_df = ensure_gdf(point_df)
    geom = first_notnull_value(point_df['geometry'])
    if geom.geom_type not in ['point', 'Point']:
      warnings.warn('左侧数据实际非点数据，将自动提取中心点进行关联')
      point_df['geometry_backup'] = point_df['geometry']
      point_df = geom_shapely2lnglat(point_df)
      point_df = geom_lnglat2shapely(point_df, delete=False)
  elif 'lng' in point_df and 'lat' in point_df:
    point_df = geom_lnglat2shapely(point_df, delete=False)
  else:
    raise KeyError('点文件中必须有经纬度或geometry')

  polygon_df = ensure_gdf(polygon_df)

  point_df = gpd.sjoin(
      point_df,
      polygon_df,
      how='left',
      predicate=predicate,
  ).drop('index_right', axis=1)

  if 'geometry_backup' in point_df:
    del point_df['geometry']
    point_df.rename(columns={'geometry_backup': 'geometry'}, inplace=True)

  if drop_geometry:
    del point_df['geometry']
  else:
    if geometry_format == 'wkb':
      point_df['geometry'] = point_df['geometry'].apply(wkb_dumps)
    elif geometry_format == 'wkt':
      point_df['geometry'] = point_df['geometry'].apply(wkt_dumps)
    elif geometry_format == 'shapely':
      pass
    else:
      raise ValueError('不支持的geometry格式')

  return pd.DataFrame(point_df)


def ensure_lnglat(lnglat) -> tuple:
  if isinstance(lnglat, (list, tuple)):
    if all([isinstance(i, (float, int)) for i in lnglat]):
      return lnglat
    else:
      raise TypeError('数据类型错误，经度和纬度都应该为数值型')
  if is_shapely(lnglat):
    geom = lnglat
  elif is_wkt(lnglat):
    geom = wkt_loads(lnglat)
  elif is_wkb(lnglat):
    geom = wkb_loads(lnglat)
  elif is_geojson(lnglat):
    geom = geojson_loads(lnglat)
  else:
    raise ValueError('未知的地理类型')
  return (geom.x, geom.y)


def distance(
    p1: (tuple, str),
    p2: (tuple, str),
    city: str = None,
    epsg_from: int = 4326,
    epsg_to: (str, int) = None):
  """
  计算两个点（经度，纬度）之间的距离，单位：米
  Args:
    p1: 点位1，经纬度或geometry
    p2: 点位2，经纬度或geometry
    city: 所在城市，用于投影
    epsg_from: epsg代码
    epsg_to: 投影epsg代码
  """
  p1, p2 = ensure_lnglat(p1), ensure_lnglat(p2)
  if not epsg_to:
    epsg_to = get_epsg(city) if city else get_epsg_by_lng([p1[0], p2[0]])
  p1 = projection_lnglat(p1, epsg_from, epsg_to)
  p2 = projection_lnglat(p2, epsg_from, epsg_to)
  return Point(p1).distance(Point(p2))



import os
import re

from setuptools import find_packages
from setuptools import setup

pwd = os.path.dirname(__file__)

with open(os.path.join(pwd, 'src', 'ricco', '__init__.py'),
          encoding='utf-8') as f:
  VERSION = (
    re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S)
    .match(f.read())
    .group(1)
  )

with open(os.path.join(pwd, 'README.md'), encoding='utf-8') as f:
  README = f.read()

setup(name='ricco',
      version=VERSION,
      description='A handy ETL&GEOM kit',
      long_description=README,
      long_description_content_type="text/markdown",
      author="Ricco Wang",
      author_email="wyk_0610@163.com",
      packages=find_packages('src'),
      package_dir={'': 'src'},
      platforms='any',
      install_requires=['numpy>=0.17',
                        'pandas>=0.24',
                        'geojson>=2.5.0',
                        'pypinyin>=0.35.0',
                        'tqdm>=4.47.0',
                        'xlrd>=1.2.0'],
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Developers',
                   'Natural Language :: English',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 3',
                   'Topic :: Software Development :: Libraries'],
      url='https://github.com/Ricco1010/ricco')



import builtins
import datetime
import json
import logging
import random
import re
import uuid
import warnings
from ast import literal_eval

import numpy as np
import pandas as pd

from ..resource.names import FirstName
from ..resource.names import LastName
from ..resource.patterns import Pattern


def ensure_list(val):
  """将标量值和Collection类型都统一转换为LIST类型"""
  if val is None:
    return []
  if isinstance(val, list):
    return val
  if isinstance(val, (set, tuple)):
    return list(val)
  return [val]


def to_json_string(string, errors='raise'):
  if not isinstance(string, (list, dict)):
    try:
      string = literal_eval(string)
    except ValueError as e:
      if errors == 'raise':
        raise ValueError(e)
      elif errors == 'coerce':
        return None
      elif errors == 'ignore':
        return string
      else:
        raise ValueError('errors传参错误')

  return json.dumps(string)


def is_ID_number(string, na=False) -> bool:
  """
  校验一个字符串是否为正确的身份证号

  :param string: 要传入的字符串
  :param na: 空值返回True还是False，默认为False
  :return:
  """
  if pd.isna(string):
    return na

  if not isinstance(string, str):
    string = str(string)
  if len(string) != 18:
    return False
  if re.match(Pattern.ID_number, string):
    return True
  else:
    return False


def relstrip(string, kwd):
  return re.sub(f'^{kwd}', '', string)


def rerstrip(string, kwd):
  return re.sub(f'{kwd}$', '', string)


def get_shortest_element(elements: list):
  """获取列表中长度最短的元素"""

  def condition(string):
    if pd.isna(string):
      return np.inf
    string = str(string)
    return len(string)

  return min(elements, key=condition)


def and_(*conditions):
  if len(conditions) < 1:
    raise Exception('Must there be one condition')
  res = conditions[0]
  for c in conditions[1:]:
    res = res & c
  return res


def or_(*conditions):
  cond = []
  for c in conditions:
    if isinstance(c, (list, tuple)):
      cond.extend(list(c))
    else:
      cond.append(c)
  if len(cond) < 1:
    raise Exception('Must there be one condition')
  res = cond[0]
  for c in cond[1:]:
    res = res | c
  return res


def all_year_old(birthday: datetime.datetime,
                 deadline: datetime.datetime = None):
  now = deadline if deadline else datetime.datetime.now()
  now_str = now.strftime('%m%d')
  birth_str = birthday.strftime('%m%d')
  if now_str >= birth_str:
    return now.year - birthday.year
  else:
    return now.year - birthday.year - 1


def first_notnull_value(series):
  for v in series:
    if pd.notna(v):
      return v
  warnings.warn('所有值均为空值')
  return None


def pinyin(word: str) -> str:
  """将中文转换为汉语拼音"""
  import pypinyin
  if isinstance(word, str):
    s = ''
    for i in pypinyin.pinyin(word, style=pypinyin.NORMAL):
      s += ''.join(i)
  else:
    raise TypeError('输入参数必须为字符串')
  return s


def is_valid_uuid(uuid_to_test, version=4):
  """判断一个字符是否是uuid"""
  try:
    uuid_obj = uuid.UUID(uuid_to_test, version=version)
  except ValueError:
    return False
  return str(uuid_obj) == uuid_to_test


def get_uuid(s):
  """针对格式错误的uuid和空白值生成新的uuid"""
  if pd.isna(s):
    return uuid.uuid4()
  elif not is_valid_uuid(s):
    return uuid.uuid4()
  else:
    return s


def per2float(string: str) -> float:
  """带有百分号的数值字符串转小数点形式的数值，
  没有百分号的返回原值"""
  if '%' in string:
    string = string.rstrip('%')
    return float(string) / 100
  else:
    return float(string)


def extract_num(string: str,
                num_type: str = 'str',
                method: str = 'list',
                join_list: bool = False,
                ignore_pct: bool = True,
                multi_warning=False):
  """
  提取字符串中的数值，默认返回所有数字组成的列表

  :param string: 输入的字符串
  :param num_type:  输出的数字类型，int/float/str，默认为str
  :param method: 结果计算方法，对结果列表求最大/最小/平均/和等，numpy方法，默认返回列表本身
  :param join_list: 是否合并列表，默认FALSE
  :param ignore_pct: 是否忽略百分号，默认True
  :param multi_warning:
  :return:
  """
  string = str(string)
  if ignore_pct:
    lis = re.findall(r"\d+\.?\d*", string)
  else:
    lis = re.findall(r"\d+\.?\d*%?", string)
  lis2 = [getattr(builtins, num_type)(per2float(i)) for i in lis]
  if len(lis2) > 0:
    if method != 'list':
      if join_list:
        raise ValueError(
            "计算结果无法join，只有在method='list'的情况下, 才能使用join_list=True")
      if multi_warning & (len(lis2) >= 2):
        warnings.warn(f'有多个值进行了{method}运算')
      res = getattr(np, method)(lis2)
    else:
      if num_type == 'str':
        res = ['{:g}'.format(float(j)) for j in lis2]
      else:
        res = lis2
      if join_list:
        res = ''.join(res)
  else:
    res = None
  return res


def to_float(string,
             rex_method: str = 'mean',
             ignore_pct: bool = False,
             multi_warning=True):
  """
  字符串转换为float
  """
  return extract_num(string,
                     num_type='float',
                     method=rex_method,
                     ignore_pct=ignore_pct,
                     multi_warning=multi_warning)


def house_type_format(x):
  """
  通过正则表达式将户型统一修改为1房，2房···5房及以上，目前只支持9室以下户型，
  其中5室及以上的类别为“5房及以上”
  """
  from ..resource import UTIL_CN_NUM

  exp = '|'.join(UTIL_CN_NUM.keys())
  pattern = f'([{exp}|\d])[室|房]'
  res = re.findall(pattern, str(x))
  if len(res) >= 1:
    res_num = res[0]
    for i in UTIL_CN_NUM:
      res_num = res_num.replace(i, UTIL_CN_NUM[i])
    if int(res_num) <= 4:
      return res_num + '房'
    else:
      return '5房及以上'
  else:
    return None


def segment(x,
            gap: (list, float, int),
            sep: str = '-',
            unit: str = '',
            bottom: str = '以下',
            top: str = '以上') -> str:
  """
  区间段划分工具

  :param x: 数值
  :param gap: 间隔，固定间隔或列表
  :param unit: 单位，末尾
  :param sep: 分隔符，中间
  :param bottom: 默认为“以下”：80米以下
  :param top: 默认为“以上”：100米以上
  :return: 区间段 'num1分隔符num2单位'：‘80-100米’
  """

  def between_list(_x, lis):
    for i in reversed(range(len(lis) - 1)):
      if _x >= lis[i]:
        return lis[i], lis[i + 1]

  x = to_float(x)
  if x is None:
    return ''
  elif isinstance(gap, list):
    gap = sorted(list(set(gap)))
    if x < gap[0]:
      return f'{gap[0]}{unit}{bottom}'
    elif x >= gap[-1]:
      return f'{gap[-1]}{unit}{top}'
    else:
      return f'{between_list(x, gap)[0]}{sep}{between_list(x, gap)[1]}{unit}'
  elif isinstance(gap, (int, float)):
    if x >= 0:
      return f'{int(x / gap) * gap}{sep}{int(x / gap) * gap + gap}{unit}'
    else:
      return f'{int(x / gap) * gap - gap}{sep}{int(x / gap) * gap}{unit}'
  else:
    raise TypeError('gap参数数据类型错误')


def fuzz_match(string: str, ss: (list, pd.Series)):
  """
  为某一字符串从某一集合中匹配相似度最高的元素

  :param string: 输入的字符串
  :param ss: 要去匹配的集合
  :return: 字符串及相似度组成的列表
  """
  from fuzzywuzzy import fuzz

  def _ratio(_s, x):
    return fuzz.ratio(_s, x), fuzz.partial_ratio(_s, x)

  max_r, max_pr, max_s = 0, 0, None
  for s in ss:
    r, pr = _ratio(s, string)
    if r > max_r:
      max_r = r
      max_pr = pr
      max_s = s
  return max_s, max_r, max_pr


def get_city_id_by_name(city: str):
  """获取城市代码"""
  from ..resource.city_id import CITY_IDS
  city_id = CITY_IDS.get(city if city.endswith('市') else f'{city}市')
  if not city_id:
    logging.warning('获取city_id失败，请补充该城市')
  return city_id


def sort_by_list(src_list, by_list, mode='sort') -> list:
  """
  根据一个列表对另一个列表进行筛选或排序，
  当mode为f/filter时，对列表进行筛选并排序，取二者交集；
  当mode为s/sort时，对列表进行排序，优先按照参照的列表排序，忽略没有的元素，
  参照列表中不存在的元素按照原始顺序排列在后
  Args:
    src_list: 要进行排序的列表
    by_list: 参照的列表
    mode: f/filter, s/sort
  """
  if mode not in ('f', 'filter', 's', 'sort'):
    raise ValueError('参数错误')

  res = [i for i in by_list if i in src_list]
  if mode in ('s', 'sort'):
    _res = [i for i in src_list if i not in res]
    res.extend(_res)
  return res


def remove_null_in_dict(dic: dict) -> dict:
  """将字典中值为空的元素删掉"""
  return {
    k: v
    for k, v in dic.items()
    if not_empty(v)
  }


def is_empty(x) -> bool:
  """判断是否为空值"""
  if isinstance(x, (list, dict, tuple)):
    return False if x else True
  if isinstance(x, (pd.DataFrame, pd.Series)):
    return x.empty
  return pd.isna(x)


def not_empty(x) -> bool:
  """判断是否非空"""
  return not is_empty(x)


def union_str(lis: list, sep='') -> (str, None):
  """连接字符串"""
  if is_empty(lis):
    return None
  lis = [i for i in lis if not_empty(i)]
  if lis:
    return sep.join(lis)
  else:
    return None


def union_list(s) -> list:
  """合并列表"""
  if is_empty(s):
    return []
  lis = s[0]
  for i in s[1:]:
    lis.extend(i)
  return lis


def eval(x: str):
  """将文本型的列表、字典等转为真正的类型"""
  return literal_eval(x) if not_empty(x) else None


def list2dict(x: list):
  """列表转为字典，key为元素的顺序"""
  if is_empty(x):
    return {}
  return {i: j for i, j in enumerate(x)}


def re_fast(pattern, string, warning=True):
  """根据正则表达式快速提取匹配到的第一个"""
  if ls := re.findall(pattern, string):
    if warning and len(ls) >= 2:
      warnings.warn('匹配到多个值，默认返回第一个')
    return ls[0]


def random_name():
  """随机生成中文名字，仅生成2或3字名字"""
  c = [random.choice(FirstName)]
  l = random.randint(1, 2)
  for i in range(l):
    c.append(random.choice(LastName))
  return ''.join(c)


def random_date():
  """获取随机日期"""
  now = datetime.datetime.now()
  tsp = int(datetime.datetime.timestamp(now))
  tsp = random.randrange(-946800000, tsp)
  return datetime.datetime.fromtimestamp(tsp).strftime("%Y%m%d")


def random_room_number(unit=False):
  """生成随机的房间号"""
  floor = random.randint(1, 17)
  room = random.randint(1, 4)
  room = str(room).zfill(2)
  if unit:
    _u = random.randint(1, 6)
    _unit = f'{_u}单元'
    return f'{_unit}{floor}{room}'
  return f'{floor}{room}'


def random_by_prob(mapping: dict):
  """根据概率生成随机值"""
  return np.random.choice(
      list(mapping.keys()),
      p=np.array(list(mapping.values())).ravel()
  )


def to_int_str(x):
  if is_empty(x):
    return
  else:
    try:
      return str(int(float(x)))
    except ValueError:
      return str(x)



import os

from tqdm import tqdm

from ..util.os import fn
from ..util.os import mkdir_2
from .extract import rdf


def split_csv(filename: str, n: int = 5, encoding: str = 'utf-8'):
  """将文件拆分为多个同名文件，放置在与文件同名文件夹下的不同Part_文件夹中"""
  dir_name = fn(os.path.basename(filename))
  abs_path = os.getcwd()
  df = rdf(filename)
  t = len(df)
  p = int(t / n)
  for i in tqdm(range(n)):
    low = i * p
    high = (i + 1) * p
    dir_name2 = 'Part_' + str(i)
    save_path = os.path.join(abs_path, dir_name, dir_name2)
    savefile = os.path.join(save_path, filename)
    mkdir_2(save_path)
    if i == n - 1:
      add = df.iloc[low:, :]
    else:
      add = df.iloc[low: high, :]
    add.to_csv(savefile, index=0, encoding=encoding)
